package de.psychometrica.lsa;



/**
 * The LSATools include some convenience methods for computing cosines and euclidean distances for semantic vectors
 * @author Wolfgang Lenhard
 *
 */
public class LSATools {

	/**
	 * returns the cosine between two texts using the cosine measure
	 * 
	 * @param source
	 *            one text
	 * @param target
	 *            the other text
	 * @return similarity between the two texts
	 */
	public static float getCosine(String source, String target) {
		return cosine(VectorStore.getInstance().getWeightedSemanticVector(source)
				.getCoordinate(), VectorStore.getInstance()
				.getWeightedSemanticVector(target).getCoordinate());
	}

	/**
	 * returns the cosine between two vectors using the cosine measure
	 * 
	 * @param vector1 the first vector
	 * @param vector2 the second vector
	 * @return similarity between the two texts
	 */
	public static float getCosine(SemanticVector vector1, SemanticVector vector2) {
		return cosine(vector1.getCoordinate(), vector2.getCoordinate());
	}

	/**
	 * returns an array of SemanticVector for a string array of texts
	 * 
	 * @param texts
	 *            Array of Strings
	 * @return array of SemanticVector for a string array of texts
	 */
	public static SemanticVector[] getVectors(String[] texts) {
		SemanticVector[] vector = new SemanticVector[texts.length];
		for (int i = 0; i < texts.length; i++)
			vector[i] = VectorStore.getInstance().getWeightedSemanticVector(
					texts[i]);
		return vector;
	}

	/**
	 * returns a SemanticVector for an given text
	 * 
	 * @param text
	 *            Simple text
	 * @return SemanticVector for an given text
	 */
	public static SemanticVector getVector(String text) {
		return VectorStore.getInstance().getWeightedSemanticVector(text);
	}


	/**
	 * return the cosine between two vectors
	 * 
	 * @param x
	 *            coordinates of vector 1
	 * @param y
	 *            coordinates of vector 2
	 * @return
	 */
	public static float cosine(float[] x, float[] y) {

		float cosine = 0f;

		double nummerator = 0f;
		double denominator1 = 0f;
		double denominator2 = 0f;

		for (int i = 0; i < x.length; i++) {
			nummerator += x[i] * y[i];
			denominator1 += x[i] * x[i];
			denominator2 += y[i] * y[i];
		}

		cosine = (float) (nummerator / (Math.sqrt(denominator1) * Math
				.sqrt(denominator2)));

		// Avoid round errors
		if (cosine > 1.0f)
			return 1.0f;
		else if (cosine < -1.0f)
			return -1.0f;
		else
			return cosine;
	}



	/** Compute the euclidean distance between two vectors within the semantic space
	 * @param vector1 First vector
	 * @param vector2 Second vector
	 * @return the distance
	 */
	public static float getDistance(SemanticVector vector1,
			SemanticVector vector2) {
		double length = .0f;
		for (int i = 0; i < vector1.getCoordinate().length; i++) {
			length += (vector1.getCoordinate(i) - vector2.getCoordinate(i))
					* (vector1.getCoordinate(i) - vector2.getCoordinate(i));
		}

		return (float) Math.sqrt(length);
	}

	/** Compute the euclidean distance between two texts within the semantic space
	 * @param source First text
	 * @param target Second text
	 * @return the distance
	 */
	public static float getDistance(String source, String target) {
		return getDistance(VectorStore.getInstance().getWeightedSemanticVector(
				source), VectorStore.getInstance()
				.getWeightedSemanticVector(target));
	}
}
