package de.psychometrica.lsa;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Hashtable;


/**This class serves for retrieving coordinates from the database file.
 * The class is used as a singleton. Please use the getInstance() to use it's methods.
 * Please set the path to the database file before doing any further actions.
 * @author Wolfgang Lenhard
 *
 */
public class VectorStore {
	private Connection con;
	private PreparedStatement getWord;
	private PreparedStatement getWordID;
	private static VectorStore connector = null;
	private static String userDir = "database";
	private static ArrayList<String> stopWords = new ArrayList<String>(1020);
	private static float[] singularValues = new float[300];
	private static HashMap<String, Long> index = new HashMap<String, Long>();

	/**
	 * Sets up a VectorStore-Object for the retrieval of coordinates
	 */
	private VectorStore() {
		try {
			System.out.print("Initializing database ...");
			Class.forName("org.h2.Driver").newInstance();
			String url = "jdbc:h2:" + userDir + File.separator
					+ "database;ACCESS_MODE_DATA=r";			
			con = DriverManager.getConnection(url);
			getWord = con
					.prepareStatement("SELECT * FROM WORDS WHERE WORD = ?");
			getWordID = con
					.prepareStatement("SELECT * FROM WORDS WHERE WORDID = ?");
			System.out.print(" deserializing indexes ...");
			deserializeIndex();
			deserializeStopwords();
			deserializeSingularValues();
			System.out.println(" done!");
		} catch (org.h2.jdbc.JdbcSQLException e) {
			System.err.println("An error occured opening the database. The database is probably already in use. Error code: " + e.getMessage());
			e.printStackTrace();
		} catch (InstantiationException e) {
			System.err.println("An error occured opening the database. Error code: " + e.getMessage());
			e.printStackTrace();
		} catch (IllegalAccessException e) {
			System.err.println("An error occured opening the database. Error code: " + e.getMessage());
			e.printStackTrace();
		} catch (ClassNotFoundException e) {
			System.err.println("An error occured opening the database. Error code: " + e.getMessage());
			e.printStackTrace();
		} catch (SQLException e) {
			System.err.println("An error occured opening the database. Error code: " + e.getMessage());
			e.printStackTrace();
		}
	}


	/**
	 * Initializes or retrieves the VectorStore
	 * @return the VectorStore
	 */
	public static VectorStore getInstance() {
		if (connector == null)
			connector = new VectorStore();

		return connector;
	}

	/**Please set up the path to the directory file before doing any computations.
	 * @param directory the directory, where all other database files reside
	 */
	public static void setPath(String directory){
		userDir = directory;
		if(!new File(userDir + File.separator + "database.h2.db").exists()){
			System.err.println("The database file does not exist under the specified path.");
		}
	}


	/**Retrieves the vector for a specific word
	 * @param word The word
	 * @return the semantic vector
	 */
	public SemanticVector getWord(String word) {
		long id = -1;
		id = index.get(word);
		if (id == -1)
			return null;

		return getWord(id, word);
	}

	/**Retrieves the vector for a specific word
	 * @param wordid The id of the word in the database
	 * @param word The word
	 * @return the semantic vector
	 */
	private SemanticVector getWord(long wordid, String word) {
		float[] coordinate = new float[300];
		float weight = 0;
		try {

			getWordID.setLong(1, wordid);
			ResultSet rs = getWordID.executeQuery();
			rs.next();
			weight = rs.getFloat(3);
			for (int i = 0; i < 300; i++) {
				coordinate[i] = rs.getFloat(4 + i);
			}

			return new SemanticVector(word, weight, coordinate);
		} catch (Exception e) {
			System.out.println(wordid + " " + word);
			e.printStackTrace();
			return null;
		}
	}


	/**
	 * Disconnects the database. 
	 */
	public void disconnect() {
		try {
			getWord.close();
			getWordID.close();
			con.close();
			VectorStore.connector = null;
		} catch (SQLException e) {
			e.printStackTrace();
		}
	}

	/**Checks, if the connection to the database is still open
	 * @return Returns true, if the connection is open, else return false.
	 */
	public boolean isConnected() {
		try {
			return !con.isClosed();
		} catch (SQLException e) {
			e.printStackTrace();
			return false;
		}
	}

	/**
	 * Deserialize index of the stop words in the database.
	 */
	@SuppressWarnings("unchecked")
	private void deserializeIndex() {
		InputStream fis = null;

		try {
			fis = new FileInputStream(userDir + File.separator + "index.dat");
			ObjectInputStream o = new ObjectInputStream(fis);
			index = (HashMap<String, Long>) o.readObject();
		} catch (IOException e) {
			System.err.println(e);
		} catch (ClassNotFoundException e) {
			System.err.println(e);
		} finally {
			try {
				fis.close();
			} catch (Exception e) {
			}
		}
	}

	/**
	 * Deserialize index of the stop words in the database.
	 */
	@SuppressWarnings("unchecked")
	private void deserializeStopwords() {
		InputStream fis = null;

		try {
			fis = new FileInputStream(userDir + File.separator
					+ "stopwords.dat");
			ObjectInputStream o = new ObjectInputStream(fis);
			stopWords = (ArrayList<String>) o.readObject();
		} catch (IOException e) {
			System.err.println(e);
		} catch (ClassNotFoundException e) {
			System.err.println(e);
		} finally {
			try {
				fis.close();
			} catch (Exception e) {
			}
		}
	}

	/**
	 * Deserialize singular values.
	 */
	private void deserializeSingularValues() {
		InputStream fis = null;

		try {
			fis = new FileInputStream(userDir + File.separator
					+ "singularValues.dat");
			ObjectInputStream o = new ObjectInputStream(fis);
			singularValues = (float[]) o.readObject();
		} catch (IOException e) {
			System.err.println(e);
		} catch (ClassNotFoundException e) {
			System.err.println(e);
		} finally {
			try {
				fis.close();
			} catch (Exception e) {
			}
		}
	}

	/**The SemanticVector for a the text is computed by adding the vectors of all single words. Afterwards, the values on the
	 * different dimensions are weighted by the singular values. To our experience, this is a favorable approach when scoring essays.
	 * @param text The text, for which the vector should be computed
	 * @return the semantic vector
	 */
	public SemanticVector getWeightedSemanticVector(String text) {
		SemanticVector vector = new SemanticVector(text);
		String[] data = text
				.replaceAll(
						"[^a-zA-ZäÄöÖüÜßÀÁÂÃÅÆÇÈÉÊÌÍÎÏÐÑÒÓÔÕÙÚÛÝÞàáâãåæçèéêëìíîïðñòóôõùúûýþÿ]+",
						" ").toLowerCase().split(" ");
		Hashtable<String, Integer> wordMap = new Hashtable<String, Integer>();
		ArrayList<SemanticVector> wordVector = new ArrayList<SemanticVector>();

		// retrieve data
		for (int i = 0; i < data.length; i++) {
			if (data[i].length() > 0) {
				try {
					Long id = -1L;
					id = index.get(data[i]);

					if ((id != null) && (id > -1)) {
						vector.incrementNumberOfWords();
						Integer number = wordMap.get(data[i]);
						if (number == null) {
							number = new Integer(1);
							wordVector.add(getWord(id, data[i]));
						} else {
							number = new Integer(number.intValue() + 1);
						}

						wordMap.put(data[i], number);
					}
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
		}

		// compute coordinates
		for (int i = 0; i < wordVector.size(); i++) {
			float powN_Weight = (float) (Math.pow(((float) wordMap
					.get(wordVector.get(i).getWord()) / vector
					.getNumberOfWords()), 0.125));
			powN_Weight = powN_Weight * wordVector.get(i).getWeight();

			for (int j = 0; j < 300; j++) {
				vector.addCoordinate(j, powN_Weight
						* wordVector.get(i).getCoordinate(j));
			}
		}

		float[] coordinates = vector.getCoordinate();
		for(int i = 0; i < coordinates.length; i++){
			coordinates[i] = coordinates[i] / singularValues[i];
		}
		vector.setCoordinate(coordinates);

		return vector;
	}

	/**The pure SemanticVector for a the text is computed by adding the vectors of all single words. In contrast to  getWeightedSemanticVector()
	 * the values on the different dimensions are not weighted by the singular values.
	 * @param text The text, for which the vector should be computed
	 * @return the semantic vector
	 */
	public SemanticVector getUnweightedSemanticVector(String text) {
		SemanticVector vector = new SemanticVector(text);
		String[] data = text
				.replaceAll(
						"[^a-zA-ZäÄöÖüÜßÀÁÂÃÅÆÇÈÉÊÌÍÎÏÐÑÒÓÔÕÙÚÛÝÞàáâãåæçèéêëìíîïðñòóôõùúûýþÿ]+",
						" ").toLowerCase().split(" ");
		Hashtable<String, Integer> wordMap = new Hashtable<String, Integer>();
		ArrayList<SemanticVector> wordVector = new ArrayList<SemanticVector>();

		// retrieve data
		for (int i = 0; i < data.length; i++) {
			if (data[i].length() > 0) {
				try {
					Long id = -1L;
					id = index.get(data[i]);

					if ((id != null) && (id > -1)) {
						vector.incrementNumberOfWords();
						Integer number = wordMap.get(data[i]);
						if (number == null) {
							number = new Integer(1);
							wordVector.add(getWord(id, data[i]));
						} else {
							number = new Integer(number.intValue() + 1);
						}

						wordMap.put(data[i], number);
					}
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
		}

		// compute coordinates
		for (int i = 0; i < wordVector.size(); i++) {
			for (int j = 0; j < 300; j++) {
				vector.addCoordinate(j, wordVector.get(i).getCoordinate(j));
			}
		}

		float[] coordinates = vector.getCoordinate();
		for(int i = 0; i < coordinates.length; i++){
			coordinates[i] = coordinates[i] / singularValues[i];
		}
		vector.setCoordinate(coordinates);

		return vector;
	}

	/**Return all unknown word in the text 
	 * @param text the text to check
	 * @return String-Array with all unknown words
	 */
	public String[] getMissingWords(String text) {
		String[] data = text
				.replaceAll(
						"[^a-zA-ZäÄöÖüÜßÀÁÂÃÅÆÇÈÉÊÌÍÎÏÐÑÒÓÔÕÙÚÛÝÞàáâãåæçèéêëìíîïðñòóôõùúûýþÿ]+",
						" ").toLowerCase().split(" ");
		ArrayList<String> wordList = new ArrayList<String>();
		for (int i = 0; i < data.length; i++) {
			if (!exists(data[i]) && (!wordList.contains(data[i])))
				wordList.add(data[i]);
		}

		String[] words = new String[wordList.size()];
		for (int i = 0; i < words.length; i++) {
			words[i] = wordList.get(i);
		}

		return words;
	}

	/**Return all known word in the text 
	 * @param text the text to check
	 * @return String-Array with all known words
	 */
	public String[] getKnownWords(String text) {
		String[] data = text
				.replaceAll(
						"[^a-zA-ZäÄöÖüÜßÀÁÂÃÅÆÇÈÉÊÌÍÎÏÐÑÒÓÔÕÙÚÛÝÞàáâãåæçèéêëìíîïðñòóôõùúûýþÿ]+",
						" ").toLowerCase().split(" ");
		ArrayList<String> wordList = new ArrayList<String>();
		for (int i = 0; i < data.length; i++) {
			if (exists(data[i]) && (!wordList.contains(data[i])))
				wordList.add(data[i]);
		}

		String[] words = new String[wordList.size()];
		for (int i = 0; i < words.length; i++) {
			words[i] = wordList.get(i);
		}

		return words;
	}

	/**Checks, if the word is in the database
	 * @param word the word to check
	 * @return true id the word exists in the database, else return false
	 */
	public boolean exists(String word) {
		return ((index.containsKey(word)) || stopWords.contains(word));
	}

	public ArrayList<String> getMissingWordsList(String text) {
		String[] data = text
				.replaceAll(
						"[^a-zA-ZäÄöÖüÜßÀÁÂÃÅÆÇÈÉÊÌÍÎÏÐÑÒÓÔÕÙÚÛÝÞàáâãåæçèéêëìíîïðñòóôõùúûýþÿ]+",
						" ").toLowerCase().split(" ");
		ArrayList<String> wordList = new ArrayList<String>();
		for (int i = 0; i < data.length; i++) {
			if (!exists(data[i]) && !wordList.contains(data[i]))
				wordList.add(data[i]);
		}

		return wordList;
	}
}
